BeginPackage["HEPMath`Compile`",
  {"HEPMath`Tensors`", "HEPMath`Lorentz`", "HEPMath`PaVe`",
   "HEPMath`LoopTools`", "HEPMath`LHAPDF`"}]

HEPPrepare::usage = "HEPPrepare[expr] prepares expression expr for numerical
evaluation. The result is of the form HEPPreparedExpression[Block[...]], and
HEPPreparedExpression has attribute HoldAll. The enclosed Block expression
evaluates to a number when all Lorentz vectors are replaced by lists containing
four real numbers. PaVe expressions are hoisted so that all loop integrals are
only calculated once. HEPCompile will automatically remove the
HEPPreparedExpression head since its only purpose is to keep to Block exression
unevaluated."

HEPPrepadedExpression::usage = "HEPPrepareExpression is the head of all
expressions returned by HEPPrepare. It has attributes {HoldAll} and its only
purpose is to keep the enclosed expression unevaluated. HEPCompile will
automatically replace HEPPreparedExpression[expr] with expr before compilation."

SetAttributes[HEPPreparedExpression, HoldAll]

HEPEvaluate::usage = "HEPEvaluate[expr] returns the evaluation of expr. The
evaluation of expr is also substituted for HEPEvaluate[expr] if it appears in
the body of a HEPCompile expression."

HEPCompile::usage = "HEPCompile[args, expr] compiles expr to bytecode (similar
to Compile), but additionally supports the functions TwoBodyDecay, CC, Pol
and MPol. It takes the options InlineEvaluation and LibraryName in addition to
all the options of Mathematica's Compile function."

HEPCompile::nolib = "Could not create library at `1`"

HEPCompiledFunction::usage = "HEPCompiledFunction is the head of expressions
returned by HEPCompile. It wraps around Mathematica's CompiledFunction
expression and evaluates the compiled function only if all arguments are
numeric."

InlineEvaluation::usage = "InlineEvaluation is an option of HEPCompile.
InlineEvaluation->patt means that the evaluation of all subexpressions
matching pattern patt are substituted before the expression is compiled. If
you want to substitute symbols remember to prevent their evaluation with
HoldPattern, i.e. InlineEvaluation->HoldPattern[sym]. Defaults to _HEPPrepare."

Options[HEPCompile] = Join[Options[Compile],
  {InlineEvaluation->(_HEPPrepare|_HEPEvaluate), LibraryName->""}]

ListPart::usage = "ListPart[list, i, j, ...] extracts the part {i,j,...}
from list. The difference to Mathematica's Part function is that ListPart
only evaluates when the first argument is a List. This is useful for plotting
and numerical integration when the function to be plotted or integrated only
evaluates to a list if called with numeric arguments."

HEPCodeGenerate::usage = "HEPCodeGenerate[cfuncs, funcnames, prefix] generates a
C module from the list of compiled functions cfuncs which provides them as a C
functions with the names in the list funcnames. It creates files called
'prefix.c', 'prefix.h' and 'prefix.py'. The return types are only relevant for
the generated python module 'prefix.py'. To generate code in a directory
different from the current directory use the option TargetDirectory. A main
program which uses the module should call the function Initialize_prefix before
any calls to the function func, and should uninitialize the module with
Uninitialize_prefix prior to exit. When compiling the module you need to link
libdl in addition to the Mathematica runtime. Unlike the Mathematica's built-in
CCodeGenerate function, the code generated by HEPCodeGenerate is capable of
calling functions from other Wolfram libraries which support the LibraryLink
interface.  This is achieved by post-processing the code generated by
CCodeGenerate. In addition, HEPCodeGenerate creates a python wrapper 'prefix.py'
which allows you to call the generated functions from python. To use the
wrapper, you have to compile the generated C module into a shared library named
'_prefix.so' and put it somewhere where your python interpreter finds it. (Also
take a look at the LibraryName option.) Mathematica tensors are automatically
translated to and from numpy arrays of the same data type."

HEPCodeGenerate::invtpe = "Invalid argument or return type `1`"

HEPCodeGenerate::wmreq =
"Callback to Mathematica detected. The generated code will probably not work."

LibraryName::usage = "LibraryName is an option of HEPCompile and
HEPCodeGenerate. If used in HEPCompile together with CompilationTarget->\"C\",
it specifies the name and path of the Wolfram library to be created. The default
is an empty string, in which case the shared library is stored internally by
Mathematica and deleted when the Mathematica session ends.

If LibraryName is used in HEPCodeGenerate it specifies the name and path to the
shared library created from the generated code. Note that HEPCodeGenerate does
not automatically compile the generated code into a shared library. You need to
do that yourself, either with the CreateLibrary function or by calling the C
compiler directly. The value of LibraryName is only used by the generated python
wrapper to locate the library.  The default value is an empty string, in which
case the name will be generated from 'prefix' argument passed to
HEPCodeGenerate. If no path is specified, the library should be in the current
directory or in a directory listed in the LD_LIBRARY_PATH environment variabe
when trying to access it through the python wrapper. Also don't forget to put
the Wolfram runtime libraries into your LD_LIBRARY_PATH."

Options[HEPCodeGenerate] = {TargetDirectory->".", LibraryName->""}


Begin["`Private`"]

Needs["HEPMath`Kinematics`"]

Needs["HEPMath`Config`"]

hold::usage = "Internal symbol equivalent to Hold."

SetAttributes[hold, HoldAll]


(* HEPEvaluate *)

HEPEvaluate[ex_] = ex


(* HEPCompile *)

selectiveEvaluate[ex_, patt_, f_] := Block[{pos, val, rul},
  pos = Position[ex, patt];
  val = f /@ Extract[ex, pos];
  rul = MapThread[Rule, {pos, val}];
  ReplacePart[ex, rul]
]

selectiveEvaluate[ex_, patt_] := selectiveEvaluate[ex, patt, Identity]

abbreviate[ex_, patt_, pref_] := Block[
  {subex = Union[Cases[ex, p:patt -> hold[p], Infinity]], rul, invrul},

  rul = Block[{$Context="HEPMath`Compile`"}, (Unique[pref] -> #)& /@ subex];
  invrul = ((HoldPattern @@ #2) -> #1)& @@@ rul;
  {rul, ex /. invrul}
]

abbreviateNoHold[ex_, patt_, pref_] := Block[{rul, res},
  {rul, res} = abbreviate[ex, patt, pref];
  {rul /. hold[x_] -> x, res}
]

makeBlock[expr_, rules_List] := Block[
  {res, eqn, vars, dvars, deqn},

  vars = First /@ rules;

  eqn = hold[Evaluate[rules]];
  eqn = Apply[Set, eqn, {2}];
  eqn = Apply[CompoundExpression, eqn, {1}];

  res = HEPPreparedExpression[Block[dvars, deqn; expr]] /.
    {dvars -> vars, deqn -> eqn};
  res //. hold[x_] -> x
]


AppendTo[$ContextPath, "HEPMath`LoopTools`Protected`"]

HEPPrepare[ex_] := Block[
  {res=ex, dummy, denrules, mprodrules, paverules, paveints, num, getint,
   mprod, eps, part},

  {denrules, res} = abbreviateNoHold[res, _Den, "den"];
  denrules = denrules /. Den[k_, m_] :> 1/HEPExpand[Sqr[k] - m^2];

  {res, denrules} = {res, denrules} /. SDot[p_, q_] :> mprod[p,q] /;
    HEPTensorSignature[p] === HEPTensorSignature[q] === {Lorentz};
  {res, denrules} = {res, denrules} /. Sqr[p_] :> mprod[p,p] /;
    HEPTensorSignature[p] === {Lorentz};
  {res, denrules} = {res, denrules} /. EpsDot[p1_, p2_, p3_, p4_] :>
      eps[p1,p2,p3,p4] /;
    HEPTensorSignature[p1] === HEPTensorSignature[p2] ===
    HEPTensorSignature[p3] === HEPTensorSignature[p4] === {Lorentz};

  {mprodrules, {res, denrules}} =
    abbreviate[{res, denrules}, _mprod, "mprod"];
  mprodrules = mprodrules /. mprod[p_, q_] :>
    p[[1]] q[[1]] - p[[2]] q[[2]] - p[[3]] q[[3]] - p[[4]] q[[4]];

  {epsrules, {res, denrules}} =
    abbreviateNoHold[{res, denrules}, _eps, "eps"];
  epsrules = epsrules /. eps[p1_, p2_, p3_, p4_] :> 
    Det[{{part[p1, 1], part[p1, 2], part[p1, 3], part[p1, 4]},
         {part[p2, 1], part[p2, 2], part[p2, 3], part[p2, 4]},
         {part[p3, 1], part[p3, 2], part[p3, 3], part[p3, 4]},
         {part[p4, 1], part[p4, 2], part[p4, 3], part[p4, 4]}}];
  epsrules = epsrules /. (sym_ -> expr_) :> (sym -> hold[expr]);
  epsrules = epsrules /. part -> Part;

  res = res /. {
    PaVe[_,_,k_][__][__] :> 0 /; k > 2,
    DPaVe[2,_,k_][__][__] :> 0 /; k > 2
  };
  res = res /. {
    PaVe[n_,mu_,k_][ii__][aa__] :>
      selectiveEvaluate[hold[getIntegrals[n][mu,aa][[getIntPos[n,k][ii]]]],
                        HoldPattern[getIntPos[__][__]]] /; k >= 0,
    DPaVe[2,mu_,k_][ii__][aa__] :>
      selectiveEvaluate[hold[getIntegrals[n][mu,aa][[getDBIntPos[k][ii]]]]
                        HoldPattern[getDBIntPos[_][__]]] /; k >= 0
  };
  {paverules, res} = abbreviate[res, HoldPattern[getIntegrals[_][__]], "pave"];

  num[getint[n_][aa__]] := N /@ getint[n][aa];
  paverules = paverules /. getIntegrals -> getint;
  paverules = selectiveEvaluate[paverules, getint[_][__], num];
  paverules = paverules /. getint -> getIntegrals;
  paverules = selectiveEvaluate[paverules, getIntegrals[_]];

  makeBlock[res, Join[mprodrules,epsrules,denrules,paverules]]
]

$ContextPath = DeleteCases[$ContextPath, "HEPMath`LoopTools`Protected`"]



$ContextPath = Join[$ContextPath,
  {"HEPMath`Kinematics`Protected`", "HEPMath`LHAPDF`Protected`"}];

SetAttributes[HEPCompile, HoldAll]

HEPCompile[args_, ex_, opt:OptionsPattern[]] := Block[
  {cex, cmd, res, libname, dummy, copts},

  cex = hold[dummy] /. dummy :> ex;

  (* inject code using the Trott-Strzebonski trick *)
  cex = cex //. {
    p:OptionValue[InlineEvaluation] :> RuleCondition[p],
    p_KineticLambda :> RuleCondition[p],
    TwoBodyDecay :> RuleCondition[cTwoBodyDecay],
    CC[Pol[k_]][sg_] :> With[{cpol=cPol, nk=(N /@ k)},
                             Conjugate[cpol[nk, sg]] /; True],
    Pol[k_][sg_] :> With[{cpol=cPol, nk=(N /@ k)}, cpol[nk, sg] /; True],
    CC[MPol[k_]][sg_] :> With[{cmpol=cMPol, nk=(N /@ k)},
                              Conjugate[cmpol[nk, sg]] /; True],
    MPol[k_][sg_] :> With[{cmpol=cMPol, nk=(N /@ k)}, cmpol[nk, sg] /; True]
  };
    
  cex = cex /. HEPCompiledFunction[f_] -> f;
  cex = cex /. CC -> Conjugate;

  If[$LHAPDFLib =!= "",
    cex = cex /. LHAPDF[id_, pid_, x_, q_] :>
      With[{nx=N[x], nq=N[q], cmd=lhapdfGet}, cmd[id, pid, nx, nq] /; True];
    cex = cex /. LHAPDFAlphaS[id_, q_] :>
      With[{nq=N[q], cmd=lhapdfAlphaS}, cmd[id, nq] /; True];

    cex = cex /. {
      LHAPDFMembers -> lhapdfMembers,
      LHAPDFOpen -> lhapdfOpen,
      LHAPDFOpenID -> lhapdfOpenID,
      LHAPDFClose -> lhapdfClose
    };
  ];

  copts = DeleteCases[Join[{opt}, Options[HEPCompile]],
    (Rule|RuleDelayed)[InlineEvaluation|LibraryName, _]];

  (* force inlining of compiled functions *)
  If[FreeQ[copts, CompilationOptions -> _],
     copts = Append[copts, CompilationOptions ->
       {"InlineCompiledFunctions"->True}]
     ,
     copts = copts /. {
      (CompilationOptions -> co_List) :>
        (CompilationOptions -> Prepend[co, "InlineCompiledFunctions"->True]),
      (CompilationOptions -> Automatic) :>
        (CompilationOptions -> {"InlineCompiledFunctions"->True})
     }
  ];
  copts = Sequence @@ copts;

  cmd = Hold[Compile[args, cex, dummy]] /. dummy -> copts;
  cmd = cmd /. HoldPattern[cex] :> RuleCondition[cex];
  cmd = cmd /. h_[s1___, Verbatim[Sequence][s2___], s3___] -> h[s1, s2, s3];
  cmd = cmd /. HEPPreparedExpression[x_] -> x;
  cmd = cmd /. hold[x_] -> x;

  res = ReleaseHold[cmd];
  libname = OptionValue[LibraryName];
  If[OptionValue[CompilationTarget] === "C" && libname =!= "",

    Quiet[DeleteFile[libname], {DeleteFile::nffil}];
    If[CopyFile[res[[9,1]], libname] === $Failed,
      Message[HEPCompile::nolib, libname],
      res = ReplacePart[res, {9,1}->libname]
    ]
  ];
  HEPCompiledFunction[res]
]

$ContextPath = DeleteCases[$ContextPath,
  "HEPMath`Kinematics`Protected`"|"HEPMath`LHAPDF`Protected`"]


(* HEPCompiledFunction *)

HEPCompiledFunction[func_][args___] := func[args] /;
  And @@ (NumericQ /@ Flatten[{args}])


(* ListPart *)

ListPart[l_List, ii__] := l[[ii]]


(* HEPCodeGenerate *)

Needs["CCodeGenerator`"]

Needs["CompiledFunctionTools`"]

ws::usage = "string pattern representing zero or more whitespaces"

handle::usage = "function that maps library names to handle variable names in
generated c module"

modname::usage = "modname holds the name of the module currently created by
codeGenerate"

firstfunc::usage = "firstfunc holds the name of the first function in the
package currently created by HEPCodeGenerate"

ws = Repeated[Whitespace, {0,1}]

processLine[s_ /; StringMatchQ[s,
  "#include"~~Whitespace~~"\"WolframRTL.h\""]] = "
#include \"WolframRTL.h\"
#include <stdio.h>
#include <dlfcn.h>
static void* HEPMath_loadLibrary(WolframLibraryData libData,
				  const char* libname)
  {
    void *lib_handle;
    int (*init)(WolframLibraryData);
    int err;
    char *error;

    lib_handle = dlopen(libname, RTLD_LAZY);
    if (!lib_handle)
    {
	fprintf(stderr, \"could not open library %s\\n\", libname);
	return NULL;
    }

    init = dlsym(lib_handle, \"WolframLibrary_initialize\");
    if ((error = dlerror()) != NULL)
    {
	fprintf(stderr, \"%s\\n\", error);
	return NULL;
    }
    err = init(libData);
    if(err)
    {
        fprintf(stderr, \"initialization of %s failed with error code %d\\n\",
		libname, err);
	return NULL;
    }

    return lib_handle;
  }

static void HEPMath_closeLibrary(void* lib_handle)
  {
    dlclose(lib_handle);
  }

static LibraryFunctionPointer HEPMath_getLibraryFunction
  (void *lib_handle, const char* fname)
  {
    LibraryFunctionPointer fn;
    char *error;


    fn = dlsym(lib_handle, fname);
    if ((error = dlerror()) != NULL)
    {
	fprintf(stderr, \"%s\\n\", error);
	return NULL;
    }

    return fn;
  }
"

processLine[s_ /; StringMatchQ[s,
  "#include"~~Whitespace~~"\""~~__~~".h\""]] := StringReplace[s,
  "#include"~~Whitespace~~"\""~~firstfunc~~".h\"" ->
  "#include \""<>modname<>".h\""]

processLine[s_ /; StringMatchQ[s,
    "static"~~Whitespace~~"WolframCompileLibrary_Functions"~~Whitespace~~
    "funStructCompile"~~ws~~";"]] :=
  s<>"\n"<>StringJoin@@(("static void* "<>handle[#]<>";\n")& /@ libs)

processLine[s_ /; StringMatchQ[s,
    "funStructCompile"~~ws~~"="~~ws~~"libData"~~ws~~"->"~~ws~~
    "compileLibraryFunctions"~~ws~~";"]] :=
  s<>"\n"<>StringJoin@@
  ((handle[#]<>" = HEPMath_loadLibrary(libData, \""<>#<>"\");\n"<>
  "if("<>handle[#]<>" == 0) return LIBRARY_FUNCTION_ERROR;\n")& /@ libs)

processLine[s_ /; StringMatchQ[s,
  "DLLEXPORT"~~Whitespace~~"int"~~Whitespace~~"Initialize_"~~__~~
  "("~~ws~~"WolframLibraryData"~~Whitespace~~"libData"~~ws~~")"]] :=
  "DLLEXPORT int Initialize_"<>modname<>"(WolframLibraryData libData)"

processLine[s_ /; StringMatchQ[s,
  "DLLEXPORT"~~Whitespace~~"void"~~Whitespace~~"Uninitialize_"~~__~~
  "("~~ws~~"WolframLibraryData"~~Whitespace~~"libData"~~ws~~")"]] :=
  "DLLEXPORT void Uninitialize_"<>modname<>"(WolframLibraryData libData)"

processLine[s_ /; StringMatchQ[s,
  (Except["="]..)~~ws~~"="~~ws~~
  "funStructCompile"~~ws~~"->"~~ws~~
  "getLibraryFunctionPointer"~~ws~~"("~~ws~~"\""~~
  (Except["\""]..)~~"\""~~ws~~","~~ws~~"\""~~
  l:(Except["\""]..)~~"\""~~ws~~")"~~ws~~";"]] :=
StringReplace[s,
  i:(Except["="]..)~~ws~~"="~~ws~~
  "funStructCompile"~~ws~~"->"~~ws~~
  "getLibraryFunctionPointer"~~ws~~"("~~ws~~"\""~~
  f:(Except["\""]..)~~"\""~~ws~~","~~ws~~"\""~~
  l:(Except["\""]..)~~"\""~~ws~~")"~~ws~~";" :>
  i<>" = HEPMath_getLibraryFunction("<>handle[l]<>", \""<>f<>"\");"]

processLine[s_ /; StringMatchQ[s,
  "initialize"~~Whitespace~~"="~~Whitespace~~"1"~~ws~~";"]] :=
  StringJoin@@(("HEPMath_closeLibrary("<>handle[#]<>");\n")& /@ libs)<>
   s

processLine[s_ /; StringMatchQ[s,
  __~~"funStructCompile"~~ws~~"->"~~ws~~"getExpressionFunctionPointer"~~__]] :=
  (Message[HEPCodeGenerate::wmreq]; s)

processLine[line_] := line


processHeaderLine[s_ /; StringMatchQ[s,
  "EXTERN_C"~~Whitespace~~"DLLEXPORT"~~Whitespace~~"int"~~Whitespace~~
  "Initialize_"~~__~~
  "("~~ws~~"WolframLibraryData"~~Whitespace~~"libData"~~ws~~");"]] :=
  "EXTERN_C DLLEXPORT int Initialize_"<>modname<>"(WolframLibraryData libData);"

processHeaderLine[s_ /; StringMatchQ[s,
  "EXTERN_C"~~Whitespace~~"DLLEXPORT"~~Whitespace~~"void"~~Whitespace~~
  "Uninitialize_"~~__~~
  "("~~ws~~"WolframLibraryData"~~Whitespace~~"libData"~~ws~~");"]] :=
  "EXTERN_C DLLEXPORT void Uninitialize_"<>modname<>
  "(WolframLibraryData libData);"

processHeaderLine[line_] := line


pyArgType[Integer] = "mint"

pyArgType[Real]    = "mreal"

pyArgType[Complex] = "mcomplex"

pyArgType[{Integer|Real|Complex, _}] = "MTensor"

pyArgType[x_] := (Message[HEPCodeGenerate::invtpe, x]; Abort[])


pyArgConversion[type:(Integer|Real|Complex), {i_}] :=
  "    carg"<>ToString[i]<>" = c_"<>pyArgType[type]<>"(arg"<>ToString[i]<>")\n"

pyArgConversion[{t:(Integer|Real|Complex), n_Integer}, {i_}] :=
  "    arg"<>ToString[i]<>", carg"<>ToString[i]<>" = mtensor(arg"<>ToString[i]<>
  ", dtype="<>pyArgType[t]<>"_t, rank="<>ToString[n]<>")\n"

pyArgConversion[{t:(Integer|Real|Complex), Verbatim[_]}, {i_}] :=
  "    arg"<>ToString[i]<>", carg"<>ToString[i]<>" = mtensor(arg"<>ToString[i]<>
  ", dtype="<>pyArgType[t]<>"_t)\n"

pyArgConversion[t_,_] := (Message[HEPCodeGenerate::invtpe, t]; Abort[])


pyWriteFunction[out_, name_, args_, rtype_] := Block[{},
  WriteString[out, "def "<>name<>"("];
  Do[WriteString[out, "arg"<>ToString[i]<>", "], {i, Length[args]}];
  WriteString[out, "):\n"];
  MapIndexed[WriteString[out, pyArgConversion[#1, #2]]&, args];  
  If[Head[rtype] === List,
    WriteString[out, "    cres = mtensor_new(lib, "<>
      pyArgType[First[rtype]]<>"_t)\n"]
    ,
    WriteString[out, "    cres = c_"<>pyArgType[rtype]<>"()\n"]
  ];
     WriteString[out, "    err = lib."<>name<>"(_libData,\n"];
  Do[WriteString[out, "        carg"<>ToString[i]<>",\n"], {i, Length[args]}];
     WriteString[out, "        ct.pointer(cres))\n"];
     WriteString[out, "    if err != 0:\n"];
     WriteString[out, "        raise RuntimeError(\n"];
     WriteString[out, "            'evaluation of function "<>name<>" failed with error code '+str(err))\n"];
  If[Head[rtype] === List,
    WriteString[out, "    res = getNumpyArray(cres)\n"];
    WriteString[out, "    mtensor_free(lib, cres)\n"];
    WriteString[out, "    return res\n"];
    ,
    WriteString[out, "    return cres.value\n"];
  ];
  WriteString[out, "\n"];
]

pyWrite[out_, prefix_, names_, args_, rtypes_, libname_] := Block[{},
       WriteString[out, "__all__ = [\n"];
  Scan[WriteString[out, "           '"<>#<>"',\n"]&, names];
       WriteString[out, "          ]\n"];

  WriteString[out, "import math\n"];
  WriteString[out, "import cmath\n"];
  WriteString[out, "import numpy as np\n"];
  WriteString[out, "import ctypes as ct\n"];
  WriteString[out, "import atexit\n"];
  WriteString[out, "import os\n"];
  WriteString[out, "from hepmath import *\n"];
  WriteString[out, "\n"];

  WriteString[out, "_dir = os.path.dirname(os.path.abspath(__file__))\n"];
  WriteString[out, "lib = ct.CDLL(os.path.join(_dir, '"<>libname<>"'))\n"];
  WriteString[out, "\n"];
  WriteString[out, "lib.Initialize_"<>prefix<>".restype = ct.c_int\n"];
  WriteString[out, "lib.Initialize_"<>prefix<>".argtypes = []\n"];
  WriteString[out, "\n"];
  WriteString[out, "lib.Uninitialize_"<>prefix<>".restype = None\n"];
  WriteString[out, "lib.Uninitialize_"<>prefix<>".argtypes = []\n"];
  WriteString[out, "\n"];
  WriteString[out, "lib.GetLibData.restype = ct.c_void_p\n"];
  WriteString[out, "lib.GetLibData.argtypes = []\n"];
  WriteString[out, "\n"];

  WriteString[out, "def _init():\n"];
  WriteString[out, "    err = lib.Initialize_"<>prefix<>"()\n"];
  WriteString[out, "    if err != 0:\n"];
  WriteString[out, "        raise RuntimeError(
            'initialization of module "<>prefix<>
            " failed with error code'+str(err))\n"];
  WriteString[out, "\n"];
  WriteString[out, "def _finish():\n"];
  WriteString[out, "    lib.Uninitialize_"<>prefix<>"()\n"];
  WriteString[out, "\n"];
  WriteString[out, "atexit.register(_finish)\n"];
  WriteString[out, "_init()\n"];
  WriteString[out, "_libData = lib.GetLibData()\n"];

  WriteString[out, "\n"];
  WriteString[out, "lib.HEPMath_MTensor_new.restype = ct.c_int\n"];
  WriteString[out, "lib.HEPMath_MTensor_new.argtypes = [c_mint, c_MTensor_p]\n"];
  WriteString[out, "\n"];
  WriteString[out, "lib.HEPMath_MTensor_free.restype = None\n"];
  WriteString[out, "lib.HEPMath_MTensor_free.argtypes = [c_MTensor]\n"];
  WriteString[out, "\n"];

  Scan[WriteString[out, "lib."<>#<>".restype = ct.c_int\n"]&, names];
  WriteString[out, "\n"];

  MapThread[Function[{n,a,r},
              WriteString[out, "lib."<>n<>".argtypes = [ct.c_void_p,\n"];
              WriteString[out, StringJoin@@
                (("    c_"<>pyArgType[#]<>",\n")& /@ a)];
             WriteString[out, "    c_"<>pyArgType[r]<>"_p]\n"];],
            {names, args, rtypes}];
  
  MapThread[pyWriteFunction[out, #1, #2, #3]&, {names, args, rtypes}];
]

codeGenerate[cfun_, name_, prefix_, dir_] := Block[
  {infile, outfile, cfile, hfile, tmpfile, line, handle,
   modname, firstfunc, libs={}, args, rtypes},

  modname = prefix<>"_"<>name;
  firstfunc = name;
  cfile = FileNameJoin[{dir,modname<>".c"}];
  hfile = FileNameJoin[{dir,modname<>".h"}];
  Quiet[CCodeGenerate[cfun, name, cfile,
        "CodeTarget"->"WolframRTL"], {CCodeGenerate::wmreq}];
  Quiet[CCodeGenerate[cfun, name, hfile,
        "CodeTarget"->"WolframRTLHeader"], {CCodeGenerate::wmreq}];

  infile = OpenRead[cfile];
  While[(line = Read[infile, String]) =!= EndOfFile,
    line = StringTrim[line];
    StringReplace[line, (Except["="]..)~~ws~~"="~~ws~~
      "funStructCompile"~~ws~~"->"~~ws~~
      "getLibraryFunctionPointer"~~ws~~"("~~ws~~"\""~~
      (Except["\""]..)~~"\""~~ws~~","~~ws~~"\""~~
      l:(Except["\""]..)~~"\""~~ws~~")"~~ws~~";" :> AppendTo[libs, l]];
  ];
  Close[infile];

  libs = Union[libs];
  MapIndexed[(handle[#1] = "LH"<>ToString[First[#2]-1])&, libs];

  tmpfile = FileNameJoin[{dir,prefix<>".c.tmp"}];
  Quiet[DeleteFile[tmpfile], {DeleteFile::nffil}];
  RenameFile[cfile, tmpfile];
  infile = OpenRead[tmpfile];
  outfile = OpenWrite[cfile];
  While[(line = Read[infile, String, NullRecords->True]) =!= EndOfFile,
    WriteString[outfile, processLine[StringTrim[line]]<>"\n"];
  ];
  Close[infile];
  Close[outfile];
  DeleteFile[tmpfile];

  tmpfile = FileNameJoin[{dir,prefix<>".h.tmp"}];
  Quiet[DeleteFile[tmpfile], {DeleteFile::nffil}];
  RenameFile[hfile, tmpfile];
  infile = OpenRead[tmpfile];
  outfile = OpenWrite[hfile];
  While[(line = Read[infile, String, NullRecords->True]) =!= EndOfFile,
    WriteString[outfile, processHeaderLine[StringTrim[line]]<>"\n"];
  ];
  Close[infile];
  Close[outfile];
  DeleteFile[tmpfile];

  cfile
]

HEPCodeGenerate[cf_List, names_List, prefix_, OptionsPattern[]] := Block[
  {cfiles, pyfile, maincfile, libname, dir, outfile, args, rtypes},

  cfuns = Replace[#, HEPCompiledFunction[f_] -> f]& /@ cf;
  dir = OptionValue[TargetDirectory];
  pyfile = FileNameJoin[{dir, prefix<>".py"}];

  cfiles = MapThread[codeGenerate[#1, #2, prefix, dir]&, {cfuns, names}];
  maincfile = FileNameJoin[{dir,prefix<>".c"}];
  outfile = OpenWrite[maincfile];
  WriteString[outfile, "#include \"WolframRTL.h\"\n"];
  WriteString[outfile, "#include \""<>prefix<>"_"<>#<>".h\"\n"]& /@ names;
  WriteString[outfile, "\nstatic WolframLibraryData libData = 0;\n"];
  WriteString[outfile, "\nstatic mbool initialize = 1;\n"];
  WriteString[outfile, "\nDLLEXPORT int Initialize_"<>prefix<>"()\n{\n"];
  WriteString[outfile, "int status=0;\n"];
  WriteString[outfile, "if (!initialize) return 0;\n"];
  WriteString[outfile, "WolframRTL_initialize(WolframLibraryVersion);\n"];
  WriteString[outfile, "libData = WolframLibraryData_new(WolframLibraryVersion);\n"];
  WriteString[outfile,
    "if((status=Initialize_"<>prefix<>"_"<>#<>"(libData)) != 0) return status;\n"]& /@
    names;
  WriteString[outfile, "initialize = 0;\n"];
  WriteString[outfile, "return 0;\n}\n"];
  WriteString[outfile, "\nDLLEXPORT void Uninitialize_"<>prefix<>"()\n{\n"];
  WriteString[outfile, "Uninitialize_"<>prefix<>"_"<>#<>"(libData);\n"]& /@ names;
  WriteString[outfile, "WolframLibraryData_free(libData);\n"];
  WriteString[outfile, "libData = 0;\n"];
  WriteString[outfile, "initialize = 1;\n"];
  WriteString[outfile, "}\n"];
  WriteString[outfile, "DLLEXPORT void* GetLibData() { return libData; }\n"];
  WriteString[outfile, "
DLLEXPORT int HEPMath_MTensor_new(mint type, MTensor* pt)
  {
    return libData->MTensor_new(type, 0, 0, pt);
  }

DLLEXPORT void HEPMath_MTensor_free(MTensor t)
  {
    libData->MTensor_free(t);
  }
"];
  Close[outfile];
  
  args = (#[[2]]& /@ cfuns) /. Verbatim[Blank][h_] -> h;
  rtypes = (ToCompiledProcedure[#][[4, 1, 1]]& /@ cfuns) /. MTensor -> List;

  outfile = OpenWrite[pyfile];
  libname = OptionValue[LibraryName];
  If[libname === "", libname = "_"<>prefix<>".so"];
  pyWrite[outfile, prefix, names, args, rtypes, libname];
  Close[outfile];

  Prepend[cfiles, maincfile]
]

HEPCodeGenerate[cf_, name_, prefix_, opt:OptionsPattern[]] :=
  HEPCodeGenerate[{cf}, {name}, prefix, opt]

HEPCodeGenerate[cf_, name_, opt:OptionsPattern[]] :=
  HEPCodeGenerate[{cf}, {name}, name, opt]



End[]

EndPackage[]


