BeginPackage["HEPMath`LoopTools`",
  {"HEPMath`Tensors`", "HEPMath`Lorentz`", "HEPMath`PaVe`"}]


Begin["`Protected`"]

getIntegrals::usage = "getIntegrals[n][mu, args] returns a list of all
n-point tensor integrals evaluated with the renormalisation scale mu and the
kinematic arguments args. These functions call LoopTools to compute the
integrals.  All arguments except n must be real numbers."

getScalarIntegral::usage = "getScalarIntegral[n][mu,k,args] returns the IRDiv^k
coefficient of the scalar n-point integral for the renormalisation scale mu and
the kinematic arguments args. This function calls LoopTools to compute the
integrals. All arguments except n must be real numbers."

getIntPos::usage = "getIntPos[n, k][inds] returns the position of the
IRDiv^k coefficient of the n-point tensor integral specified by the indices
inds in the list returned by getIntegrals[n][mu,...]."

getDBIntPos::usage = "getDBIntPos[k][inds] returns the position of the IRDiv^k
coefficient of the derivatives of the two-point tensor integrals specified by
the indices inds in the list returned by getIntegrals[2][mu,...]"

getIntPos::invidx = getDBIntPos::invidx = "Unknown tensor integral `1`"

End[]


Begin["`Private`"]

Needs["HEPMath`Config`"]

AppendTo[$ContextPath, "HEPMath`LoopTools`Protected`"]

SetAttributes[libSet, HoldFirst]

libSet[var_, func_, args_, res_] := If[$LoopToolsLib =!= "",
  var = LibraryFunctionLoad[$LoopToolsLib, func, args, res],
  var := (Message[$LoopToolsLib::noif]; Abort[])
]


(* one-point integrals *)

libSet[getIntegrals[1], "GetAIntegrals", {Real, Real}, {Complex, 1}]

libSet[getScalarIntegral[1], "GetA0", {Real, Integer, Real}, Complex]


getIntPos[1,l_][0] = 0+1+l;
getIntPos[1,l_][0,0] = 3+1+l;
getIntPos[1,l_][ii___] := (Message[getIntPos::invidx, {ii}]; Abort[])


(* two-point integrals *)

libSet[getIntegrals[2], "GetBIntegrals",
  {Real, Real, Real, Real}, {Complex, 1}]

libSet[getScalarIntegral[2], "GetB0",
  {Real, Integer, Real, Real, Real}, Complex]

getIntPos[2,l_][0]     =  0+1+l;
getIntPos[2,l_][1]     =  3+1+l;
getIntPos[2,l_][0,0]   =  6+1+l;
getIntPos[2,l_][1,1]   =  9+1+l;
getIntPos[2,l_][0,0,1] = 12+1+l;
getIntPos[2,l_][1,1,1] = 15+1+l;
getIntPos[2,l_][ii___] := (Message[getIntPos::invidx, {ii}]; Abort[])

getDBIntPos[l_][0]     = 18+1+l;
getDBIntPos[l_][1]     = 21+1+l;
getDBIntPos[l_][0,0]   = 24+1+l;
getDBIntPos[l_][1,1]   = 27+1+l;
getDBIntPos[l_][0,0,1] = 30+1+l;
getDBIntPos[l_][ii___] := (Message[getDBIntPos::invidx, {ii}]; Abort[])

If[$LoopToolsLib =!= "",
  DPaVe[2,mu_?nrealQ,k_Integer][ii__][psq_?nrealQ,m0sq_?nrealQ,m1sq_?nrealQ] :=
    If[k > 2, 0.0,
       Part[getIntegrals[2] @@ (N /@ {mu, psq, m0sq, m1sq}),
            getDBIntPos[k][ii]]] /; k >= 0
]


(* three-point integrals *)

libSet[getIntegrals[3], "GetCIntegrals",
  {Real, Real, Real, Real, Real, Real, Real}, {Complex, 1}]

libSet[getScalarIntegral[3], "GetC0",
  {Real, Integer, Real, Real, Real, Real, Real, Real}, Complex]

getIntPos[3,l_][0] = 0+1+l;
getIntPos[3,l_][1] = 3+1+l;
getIntPos[3,l_][2] = 6+1+l;
getIntPos[3,l_][0,0] = 9+1+l;
getIntPos[3,l_][1,1] = 12+1+l;
getIntPos[3,l_][1,2] = 15+1+l;
getIntPos[3,l_][2,2] = 18+1+l;
getIntPos[3,l_][0,0,1] = 21+1+l;
getIntPos[3,l_][0,0,2] = 24+1+l;
getIntPos[3,l_][1,1,1] = 27+1+l;
getIntPos[3,l_][1,1,2] = 30+1+l;
getIntPos[3,l_][1,2,2] = 33+1+l;
getIntPos[3,l_][2,2,2] = 36+1+l;
getIntPos[3,l_][0,0,0,0] = 39+1+l;
getIntPos[3,l_][0,0,1,1] = 42+1+l;
getIntPos[3,l_][0,0,1,2] = 45+1+l;
getIntPos[3,l_][0,0,2,2] = 48+1+l;
getIntPos[3,l_][1,1,1,1] = 51+1+l;
getIntPos[3,l_][1,1,1,2] = 54+1+l;
getIntPos[3,l_][1,1,2,2] = 57+1+l;
getIntPos[3,l_][1,2,2,2] = 60+1+l;
getIntPos[3,l_][2,2,2,2] = 63+1+l;
getIntPos[3,l_][ii___] := (Message[getIntPos::invidx, {ii}]; Abort[])


(* four-point integrals *)

libSet[getIntegrals[4], "GetDIntegrals",
  {Real, Real, Real, Real, Real, Real, Real, Real, Real, Real, Real},
  {Complex, 1}]

libSet[getScalarIntegral[4], "GetD0",
  {Real, Integer, Real, Real, Real, Real, Real, Real, Real, Real, Real, Real},
  Complex]

getIntPos[4,l_][0] = 0+1+l;
getIntPos[4,l_][1] = 3+1+l;
getIntPos[4,l_][2] = 6+1+l;
getIntPos[4,l_][3] = 9+1+l;
getIntPos[4,l_][0,0] = 12+1+l;
getIntPos[4,l_][1,1] = 15+1+l;
getIntPos[4,l_][1,2] = 18+1+l;
getIntPos[4,l_][1,3] = 21+1+l;
getIntPos[4,l_][2,2] = 24+1+l;
getIntPos[4,l_][2,3] = 27+1+l;
getIntPos[4,l_][3,3] = 30+1+l;
getIntPos[4,l_][0,0,1] = 33+1+l;
getIntPos[4,l_][0,0,2] = 36+1+l;
getIntPos[4,l_][0,0,3] = 39+1+l;
getIntPos[4,l_][1,1,1] = 42+1+l;
getIntPos[4,l_][1,1,2] = 45+1+l;
getIntPos[4,l_][1,1,3] = 48+1+l;
getIntPos[4,l_][1,2,2] = 51+1+l;
getIntPos[4,l_][1,2,3] = 54+1+l;
getIntPos[4,l_][1,3,3] = 57+1+l;
getIntPos[4,l_][2,2,2] = 60+1+l;
getIntPos[4,l_][2,2,3] = 63+1+l;
getIntPos[4,l_][2,3,3] = 66+1+l;
getIntPos[4,l_][3,3,3] = 69+1+l;
getIntPos[4,l_][0,0,0,0] = 72+1+l;
getIntPos[4,l_][0,0,1,1] = 75+1+l;
getIntPos[4,l_][0,0,1,2] = 78+1+l;
getIntPos[4,l_][0,0,1,3] = 81+1+l;
getIntPos[4,l_][0,0,2,2] = 84+1+l;
getIntPos[4,l_][0,0,2,3] = 87+1+l;
getIntPos[4,l_][0,0,3,3] = 90+1+l;
getIntPos[4,l_][1,1,1,1] = 93+1+l;
getIntPos[4,l_][1,1,1,2] = 96+1+l;
getIntPos[4,l_][1,1,1,3] = 99+1+l;
getIntPos[4,l_][1,1,2,2] = 102+1+l;
getIntPos[4,l_][1,1,2,3] = 105+1+l;
getIntPos[4,l_][1,1,3,3] = 108+1+l;
getIntPos[4,l_][1,2,2,2] = 111+1+l;
getIntPos[4,l_][1,2,2,3] = 114+1+l;
getIntPos[4,l_][1,2,3,3] = 117+1+l;
getIntPos[4,l_][1,3,3,3] = 120+1+l;
getIntPos[4,l_][2,2,2,2] = 123+1+l;
getIntPos[4,l_][2,2,2,3] = 126+1+l;
getIntPos[4,l_][2,2,3,3] = 129+1+l;
getIntPos[4,l_][2,3,3,3] = 132+1+l;
getIntPos[4,l_][3,3,3,3] = 135+1+l;
getIntPos[4,l_][0,0,0,0,1] = 138+1+l;
getIntPos[4,l_][0,0,0,0,2] = 141+1+l;
getIntPos[4,l_][0,0,0,0,3] = 144+1+l;
getIntPos[4,l_][0,0,1,1,1] = 147+1+l;
getIntPos[4,l_][0,0,1,1,2] = 150+1+l;
getIntPos[4,l_][0,0,1,1,3] = 153+1+l;
getIntPos[4,l_][0,0,1,2,2] = 156+1+l;
getIntPos[4,l_][0,0,1,2,3] = 159+1+l;
getIntPos[4,l_][0,0,1,3,3] = 162+1+l;
getIntPos[4,l_][0,0,2,2,2] = 165+1+l;
getIntPos[4,l_][0,0,2,2,3] = 168+1+l;
getIntPos[4,l_][0,0,2,3,3] = 171+1+l;
getIntPos[4,l_][0,0,3,3,3] = 174+1+l;
getIntPos[4,l_][1,1,1,1,1] = 177+1+l;
getIntPos[4,l_][1,1,1,1,2] = 180+1+l;
getIntPos[4,l_][1,1,1,1,3] = 183+1+l;
getIntPos[4,l_][1,1,1,2,2] = 186+1+l;
getIntPos[4,l_][1,1,1,2,3] = 189+1+l;
getIntPos[4,l_][1,1,1,3,3] = 192+1+l;
getIntPos[4,l_][1,1,2,2,2] = 195+1+l;
getIntPos[4,l_][1,1,2,2,3] = 198+1+l;
getIntPos[4,l_][1,1,2,3,3] = 201+1+l;
getIntPos[4,l_][1,1,3,3,3] = 204+1+l;
getIntPos[4,l_][1,2,2,2,2] = 207+1+l;
getIntPos[4,l_][1,2,2,2,3] = 210+1+l;
getIntPos[4,l_][1,2,2,3,3] = 213+1+l;
getIntPos[4,l_][1,2,3,3,3] = 216+1+l;
getIntPos[4,l_][1,3,3,3,3] = 219+1+l;
getIntPos[4,l_][2,2,2,2,2] = 222+1+l;
getIntPos[4,l_][2,2,2,2,3] = 225+1+l;
getIntPos[4,l_][2,2,2,3,3] = 228+1+l;
getIntPos[4,l_][2,2,3,3,3] = 231+1+l;
getIntPos[4,l_][2,3,3,3,3] = 234+1+l;
getIntPos[4,l_][3,3,3,3,3] = 237+1+l;
getIntPos[4,l_][ii___] := (Message[getIntPos::invidx, {ii}]; Abort[])


(* numerical evaluation of PaVe *)

nrealQ[x_] := Element[x, Reals] === True

If[$LoopToolsLib =!= "",
  PaVe[n:(1|2|3|4), mu_?nrealQ, k_Integer][0][args__?nrealQ] :=
    If[k > 2, 0.0,
       getScalarIntegral[n] @@ (N /@ {mu, k, args})] /; k >= 0;

  PaVe[n:(1|2|3|4), mu_?nrealQ, k_Integer][ii__Integer][args__?nrealQ] :=
    If[k > 2, 0.0,
       Part[getIntegrals[n] @@ (N /@ {mu, args}),
            getIntPos[n, k][ii]]] /; k >= 0;
]


$ContextPath = DeleteCases[$ContextPath, "HEPMath`LoopTools`Protected`"]


End[]

EndPackage[]
