#include "WolframLibrary.h"
#include <map>
#include <vector>
#include "LHAPDF/LHAPDF.h"

static int initialize = 1;

static std::map<int, LHAPDF::PDF*> pdf;

EXTERN_C DLLEXPORT int numMembers(WolframLibraryData libData,
				  mint Argc, MArgument *Args, MArgument Res)
{
    if(Argc < 1) return LIBRARY_FUNCTION_ERROR;

    char* name = MArgument_getUTF8String(Args[0]);
    mint n;
    try {
	LHAPDF::PDFSet set(name);
	n = mint(set.size());
    } catch(const LHAPDF::ReadError& e) {
	libData->UTF8String_disown(name);
	return LIBRARY_FUNCTION_ERROR;
    }

    MArgument_setInteger(Res, n);
    libData->UTF8String_disown(name);

    return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int openPDF(WolframLibraryData libData,
			       mint Argc, MArgument *Args, MArgument Res)
{
    if(Argc < 2) return LIBRARY_FUNCTION_ERROR;    

    char* name = MArgument_getUTF8String(Args[0]);
    int mem = MArgument_getInteger(Args[1]);

    LHAPDF::PDF* p = 0;
    try {
	p = LHAPDF::mkPDF(name, mem);
    } catch(const LHAPDF::ReadError& e) {
	libData->UTF8String_disown(name);
	return LIBRARY_FUNCTION_ERROR;
    } catch(const LHAPDF::UserError& e) {
	libData->UTF8String_disown(name);
	return LIBRARY_FUNCTION_ERROR;
    }
    if(p == 0) {
	libData->UTF8String_disown(name);
	return LIBRARY_FUNCTION_ERROR;
    }

    int id = p->lhapdfID();
    pdf.insert( std::pair<int, LHAPDF::PDF*>(id, p) );

    MArgument_setInteger(Res, id);
    libData->UTF8String_disown(name);

    return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int openID(WolframLibraryData libData,
			      mint Argc, MArgument *Args, MArgument Res)
{
    if(Argc < 1) return LIBRARY_FUNCTION_ERROR;    

    int idarg = MArgument_getInteger(Args[0]);

    std::pair<std::string, int> name_mem = LHAPDF::lookupPDF(idarg);
    std::string name = name_mem.first;
    int mem = name_mem.second;
    if(mem < 0)
	return LIBRARY_FUNCTION_ERROR;

    LHAPDF::PDF* p = 0;
    try {
	p = LHAPDF::mkPDF(name, mem);
    } catch(const LHAPDF::ReadError& e) {
	return LIBRARY_FUNCTION_ERROR;
    } catch(const LHAPDF::UserError& e) {
	return LIBRARY_FUNCTION_ERROR;
    }
    if(p == 0) {
	return LIBRARY_FUNCTION_ERROR;
    }

    int id = p->lhapdfID();
    pdf.insert( std::pair<int, LHAPDF::PDF*>(id, p) );

    MArgument_setInteger(Res, id);
    return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int closePDF(WolframLibraryData libData,
				mint Argc, MArgument *Args, MArgument Res)
{
    if(Argc < 1) return LIBRARY_FUNCTION_ERROR;
    int id = MArgument_getInteger(Args[0]);

    std::map<int, LHAPDF::PDF*>::const_iterator p = pdf.find(id);
    if(p != pdf.end())
	delete p->second;
    pdf.erase(id);

    MArgument_setInteger(Res, 0);
    return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int getFlavors(WolframLibraryData libData,
				  mint Argc, MArgument *Args, MArgument Res)
{
    if(Argc < 1) return LIBRARY_FUNCTION_ERROR;
    int id = MArgument_getInteger(Args[0]);

    std::map<int, LHAPDF::PDF*>::const_iterator p = pdf.find(id);
    if(p == pdf.end())
	return LIBRARY_FUNCTION_ERROR;

    MTensor res;
    int err = LIBRARY_NO_ERROR;
    mint dims[1];
    dims[0] = p->second->flavors().size();

    err = libData->MTensor_new(MType_Integer, 1, dims, &res);
    for(mint i=1; i <= dims[0] && !err; ++i) {
	err = libData->MTensor_setInteger(res, &i, p->second->flavors()[i]);
    }

    MArgument_setMTensor(Res, res);
    return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int getPDF(WolframLibraryData libData,
			      mint Argc, MArgument *Args, MArgument Res)
{
    if(Argc < 4) return LIBRARY_FUNCTION_ERROR;
    int id = MArgument_getInteger(Args[0]);
    int pid = MArgument_getInteger(Args[1]);
    mreal x = MArgument_getReal(Args[2]);
    mreal q = MArgument_getReal(Args[3]);

    std::map<int, LHAPDF::PDF*>::const_iterator p = pdf.find(id);
    if(p == pdf.end())
	return LIBRARY_FUNCTION_ERROR;
    
    MArgument_setReal(Res, p->second->xfxQ2(pid, x, q*q));
    return LIBRARY_NO_ERROR;
}

EXTERN_C DLLEXPORT int getAlphaS(WolframLibraryData libData,
			      mint Argc, MArgument *Args, MArgument Res)
{
    if(Argc < 2) return LIBRARY_FUNCTION_ERROR;
    int id = MArgument_getInteger(Args[0]);
    mreal q = MArgument_getReal(Args[1]);

    std::map<int, LHAPDF::PDF*>::const_iterator p = pdf.find(id);
    if(p == pdf.end())
	return LIBRARY_FUNCTION_ERROR;
    
    MArgument_setReal(Res, p->second->alphasQ2(q*q));
    return LIBRARY_NO_ERROR;
}


EXTERN_C DLLEXPORT mint WolframLibrary_getVersion()
{
    return WolframLibraryVersion;
}

EXTERN_C DLLEXPORT int WolframLibrary_initialize(WolframLibraryData libData)
{
    if(initialize == 1) {
	pdf.clear();
	initialize = 0;
    }

    return 0;
}

EXTERN_C DLLEXPORT void WolframLibrary_uninitialize(WolframLibraryData libData)
{
    if(initialize == 0) {
	for(std::map<int, LHAPDF::PDF*>::iterator p = pdf.begin();
	    p != pdf.end(); ++p) {
	    
	    delete p->second;
	}
	pdf.clear();

	initialize = 1;
    }

    return;
}


